@@ -26,15 +26,14 @@ special_data_methods = [
     'p_electron_fluctuation',
     'electron_acceptance',
     'photon_acceptance',
-    's1_acceptance',
-    's2_acceptance',
     'penning_quenching_eff'
 ]
 
 data_methods = (
     special_data_methods
-    + ['energy_spectrum', 'work', 'double_pe_fraction'])
-hidden_vars_per_quanta = 'detection_eff gain_mean gain_std'.split()
+    + ['energy_spectrum', 'work', 'double_pe_fraction',
+       's1_acceptance', 's2_acceptance'])
+hidden_vars_per_quanta = ['detection_eff', 'gain_mean', 'gain_std']
 for _qn in quanta_types:
     data_methods += [_qn + '_' + x for x in hidden_vars_per_quanta]
 
@@ -50,80 +49,101 @@ class LXeSource(fd.Source):
         'photon_produced',
         'electron_produced')
 
-    # tuple with columns needed from data
-    # I guess we don't really need x y z by default, but they are just so nice
-    # we should keep them around regardless.
-    extra_needed_columns = tuple(['x', 'y', 'z', 'r', 'theta'])
-
     # Whether or not to simulate overdispersion in electron/photon split
     # (e.g. due to non-binomial recombination fluctuation)
     do_pel_fluct: bool
 
-    tpc_radius = 47.9   # cm
-    tpc_length = 97.6   # cm
+    # The fiducial volume bounds for a cylindrical volume
+    # default to full (2t) XENON1T dimensions
+    fv_radius = 47.9   # cm
+    fv_high = 0  # cm
+    fv_low = -97.6  # cm
+
     drift_velocity = 1.335 * 1e-4   # cm/ns
 
-    # Uniform timestamps between 2016-09 and 2017-09
-    t_start = pd.to_datetime('2016-09-13T12:00:00')
-    t_stop = pd.to_datetime('2017-09-13T12:00:00')
+    # The default boundaries are at points where the WIMP wind is at its
+    # average speed.
+    # This will then also be true at the midpoint of these times.
+    t_start = pd.to_datetime('2019-09-01T08:28:00')
+    t_stop = pd.to_datetime('2020-09-01T08:28:00')
 
     # Spatial rate multiplier histogram
     # Multihist Histdd object to lookup space dependent rate multipliers
     # The histogram must have 'axis_names' set to either
     # ['r', 'theta', 'z'] or ['x', 'y', 'z']
-    # The histogram must be normalized to the histogram mean
+    # Must be events per bin histogram, not pdf
     spatial_rate_hist = None
+    spatial_rate_bin_volumes = None
+
+    # Whether to check efficiencies and acceptances are positive at
+    # the observed events.
+    # This is recommended, but you'll have to turn it off if your
+    # likelihood includes regions where only anomalous sources make events.
+    check_efficiencies = True
+    check_acceptances = True
 
     def __init__(self, *args, **kwargs):
+        # Deprecate tpc_radius and tpc_length
+        if hasattr(self, 'tpc_radius') or hasattr(self, 'tpc_length'):
+            raise DeprecationWarning("You've set either 'tpc_radius' or"
+                " 'tpc_length', these are deprecated. Use 'fv_radius',"
+                " 'fv_high' and 'fv_low' to denote the boundaries of the"
+                " detector.")
+
+        assert self.fv_low < self.fv_high, \
+            f"fv_low ({self.fv_low}) not less then fv_high ({self.fv_high})"
+
         # Check validity of spatial rate hist
         if self.spatial_rate_hist is not None:
+            assert self.spatial_rate_bin_volumes is not None, \
+                "Must give bin volumes as well"
             assert isinstance(self.spatial_rate_hist, Histdd), \
                 "spatial_rate_hist needs to be a multihist Histdd object"
-            # The histogram mean needs to be normalized to one, meaning the
-            # histogram contains nbins/dim**ndims counts. This ensures we can
-            # use tf.ones is no hist is defined.
-            h = self.spatial_rate_hist.histogram
-            assert np.allclose(h.mean() * h.size, self.spatial_rate_hist.n), \
-                "spatial_rate_hist needs to be normalized to histogram mean"
             # Check histogram dimensions
             axes = self.spatial_rate_hist.axis_names
             assert axes == ['r', 'theta', 'z'] or axes == ['x', 'y', 'z'], \
                 ("axis_names of spatial_rate_hist must be either "
                  "or ['r', 'theta', 'z'] or ['x', 'y', 'z']")
             self.spatial_rate_hist_dims = axes
+
+            # Correctly scale the events/bin histogram E to make the pdf R
+            # histogram, taking into account (non uniform) bin volumes. This
+            # ensures we don't need to modify mu_before_efficiencies.
+            # R = E / bv
+            # R_norm = (E / sum E) / (bv / sum bv)
+            # R_norm = (E / bv) * (sum bv / sum E)
+            bv = self.spatial_rate_bin_volumes
+            E = self.spatial_rate_hist.histogram
+            R_norm = (E / bv) * (bv.sum() / E.sum())
+
+            self.spatial_rate_pdf = self.spatial_rate_hist.similar_blank_hist()
+            self.spatial_rate_pdf.histogram = R_norm
+
         # Init rest of Source, this must be done after any checks on
         # spatial_rate_hist since it calls _populate_tensor_cache as well
         super().__init__(*args, **kwargs)
 
-    def _populate_tensor_cache(self):
-        super()._populate_tensor_cache()
+    def add_extra_columns(self, d):
+        super().add_extra_columns(d)
         if self.spatial_rate_hist is not None:
             # Setup tensor of histogram for lookup
-            positions = self.data[self.spatial_rate_hist_dims].values.T
-            v = self.spatial_rate_hist.lookup(*positions)
-            spatial_rate_tensor = tf.convert_to_tensor(v,
-                                                       dtype=fd.float_type())
-            self.spatial_rate_tensor = tf.reshape(spatial_rate_tensor,
-                                                  [self.n_batches, -1])
+            positions = d[self.spatial_rate_hist_dims].values.T
+            v = self.spatial_rate_pdf.lookup(*positions)
+            d['spatial_rate_multiplier'] = v
         else:
-            # If no hist defined, set uniform response
-            self.spatial_rate_tensor = tf.ones([self.n_batches,
-                                                self.batch_size],
-                                               dtype=fd.float_type())
+            d['spatial_rate_multiplier'] = 1.
 
     ##
     # Model functions (data_methods)
     ##
 
     # Single constant energy spectrum
-    def energy_spectrum(self, i_batch):
+    def energy_spectrum(self, spatial_rate_multiplier):
         # Lookup the energy spectrum
         es, rs = self._single_spectrum()
-        # Lookup the spatial rate multiplier and scale the spectrum with it
-        sr = self.spatial_rate_mult(i_batch)
-        n = i_batch.shape[0]
-        return (fd.repeat(es[o, :], n, axis=0),
-                fd.repeat(rs[o, :], n, axis=0) * sr[:, o])
+        sr = spatial_rate_multiplier
+        return (fd.repeat(es[o, :], self.batch_size, axis=0),
+                fd.repeat(rs[o, :], self.batch_size, axis=0) * sr[:, o])
 
     work = 13.7e-3
 
@@ -242,20 +262,16 @@ class LXeSource(fd.Source):
     def random_truth_observables(self, n_events):
         """Return dictionary with x, y, z, r, theta, drift_time
         and event_time randomly drawn.
-        S1 and S2 placeholder values are added for set_data.
         Takes into account spatial rate multiplier of the source.
         """
         data = dict()
-        # Add fake s1, s2 necessary for set_data to succeed
-        # TODO: check if we still need this...
-        data['s1'] = 1
-        data['s2'] = 100
 
         if self.spatial_rate_hist is None:
             # Draw uniform position
-            data['r'] = (np.random.rand(n_events) * self.tpc_radius**2)**0.5
+            data['r'] = (np.random.rand(n_events) * self.fv_radius**2)**0.5
             data['theta'] = np.random.uniform(0, 2*np.pi, size=n_events)
-            data['z'] = - np.random.rand(n_events) * self.tpc_length
+            data['z'] = np.random.uniform(self.fv_low, self.fv_high,
+                                          size=n_events)
             data['x'], data['y'] = fd.pol_to_cart(data['r'], data['theta'])
         elif self.spatial_rate_hist_dims == ['r', 'theta', 'z']:
             # Spatial response in cylindrical coords
@@ -274,9 +290,9 @@ class LXeSource(fd.Source):
 
         # Draw uniform time
         data['event_time'] = np.random.uniform(
-            pd.Timestamp(self.t_start).value,
-            pd.Timestamp(self.t_stop).value,
-            size=n_events).astype('float32')
+            self.t_start.value,
+            self.t_stop.value,
+            size=n_events)
         return data
 
     ##
@@ -374,7 +390,7 @@ class LXeSource(fd.Source):
         ndet = self.domain(quanta_type + '_detected', data_tensor)
 
         observed = self._fetch(
-            signal_name[quanta_type], data_tensor=data_tensor)[:, o]
+            signal_name[quanta_type], data_tensor=data_tensor)
 
         # Lookup signal gain mean and std per detected quanta
         mean_per_q = self.gimme(quanta_type + '_gain_mean',
@@ -396,12 +412,11 @@ class LXeSource(fd.Source):
         # add offset to std to avoid NaNs from norm.pdf if std = 0
         result = tfp.distributions.Normal(
                 loc=mean, scale=std + 1e-10
-            ).prob(observed)
+            ).prob(observed[:, o])
 
         # Add detection/selection efficiency
         result *= self.gimme(signal_name[quanta_type] + '_acceptance',
-                             bonus_arg=observed,
-                             data_tensor=data_tensor, ptensor=ptensor)
+                             data_tensor=data_tensor, ptensor=ptensor)[:, o]
         return result
 
     @staticmethod
@@ -431,6 +446,18 @@ class LXeSource(fd.Source):
                 else self.min_s2_electrons_detected,
                 None)
 
+    def _check_data(self):
+        super()._check_data()
+        if not self.check_acceptances:
+            return
+        for sn in signal_name.values():
+            s_acc = self.gimme(sn + '_acceptance',
+                               data_tensor=None, ptensor=None, numpy_out=True)
+            if np.any(s_acc <= 0):
+                raise ValueError(f"Found event with non-positive {sn} "
+                                 f"acceptance: did you apply and configure "
+                                 "your cuts correctly?")
+
     def _annotate(self, _skip_bounds_computation=False):
         d = self.data
 
@@ -440,10 +467,17 @@ class LXeSource(fd.Source):
             for parname in hidden_vars_per_quanta:
                 fname = qn + '_' + parname
                 try:
-                    d[fname] = self.gimme(fname, data_tensor=None, ptensor=None, numpy_out=True)
+                    d[fname] = self.gimme(fname, data_tensor=None,
+                                          ptensor=None, numpy_out=True)
                 except Exception:
                     print(fname)
                     raise
+            if (self.check_efficiencies
+                    and np.any(d[qn + '_detection_eff'].values <= 0)):
+                raise ValueError(f"Found event with non-positive {qn} "
+                                 "detection efficiency: did you apply and "
+                                 "configure your cuts correctly?")
+
         d['double_pe_fraction'] = self.gimme('double_pe_fraction',
                                              data_tensor=None, ptensor=None,
                                              numpy_out=True)
@@ -526,10 +560,8 @@ class LXeSource(fd.Source):
                 # For detected quanta the MLE is quite accurate
                 # (since fluctuations are tiny)
                 # so let's just use the relative error on the MLE
-                d[qn + '_detected_' + bound] = stats.norm.ppf(
-                    stats.norm.cdf(sign * self.max_sigma),
-                    loc=n,
-                    scale=scale,
+                d[qn + '_detected_' + bound] = (
+                    n + sign * self.max_sigma * scale
                 ).round().clip(*self._q_det_clip_range(qn)).astype(np.int)
 
                 # For produced quanta, it is trickier, since the number
@@ -537,10 +569,10 @@ class LXeSource(fd.Source):
                 # TODO: where did this derivation come from again?
                 # TODO: maybe do a second bound based on CES
                 q = 1 / eff
-                d[qn + '_produced_' + bound] = stats.norm.ppf(
-                    stats.norm.cdf(sign * self.max_sigma),
-                    loc=n_prod_mle,
-                    scale=(q + (q**2 + 4 * n_prod_mle * q)**0.5)/2
+                _loc = n_prod_mle
+                _std = (q + (q**2 + 4 * n_prod_mle * q)**0.5)/2
+                d[qn + '_produced_' + bound] = (
+                    _loc + sign * self.max_sigma * _std
                 ).round().clip(*self._q_det_clip_range(qn)).astype(np.int)
 
             # Finally, round the detected MLEs
@@ -567,8 +599,11 @@ class LXeSource(fd.Source):
 
         if self.do_pel_fluct:
             d['p_el_fluct'] = gimme('p_electron_fluctuation', d['nq'].values)
-            d['p_el_actual'] = stats.beta.rvs(
-                *fd.beta_params(d['p_el_mean'], d['p_el_fluct']))
+            d['p_el_fluct'] = np.clip(d['p_el_fluct'].values,
+                                      fd.MIN_FLUCTUATION_P,
+                                      1.)
+            d['p_el_actual'] = 1. - stats.beta.rvs(
+                *fd.beta_params(1. - d['p_el_mean'], d['p_el_fluct']))
         else:
             d['p_el_fluct'] = 0.
             d['p_el_actual'] = d['p_el_mean']
@@ -599,7 +634,7 @@ class LXeSource(fd.Source):
         for q in quanta_types:
             acceptance *= gimme(q + '_acceptance', d[q + '_detected'].values)
             sn = signal_name[q]
-            acceptance *= gimme(sn + '_acceptance', d[sn].values)
+            acceptance *= gimme(sn + '_acceptance')
         return d.iloc[np.random.rand(len(d)) < acceptance].copy()
 
     def mu_before_efficiencies(self, **params):
@@ -612,10 +647,6 @@ class LXeSource(fd.Source):
     def _single_spectrum(self):
         raise NotImplementedError
 
-    def spatial_rate_mult(self, i_batch):
-        batch = tf.dtypes.cast(i_batch[0], dtype=fd.int_type())
-        return self.spatial_rate_tensor[batch]
-
 
 @export
 class ERSource(LXeSource):
@@ -769,6 +800,7 @@ class NRSource(LXeSource):
     def rate_nq(self, nq_1d, data_tensor, ptensor):
         # (n_events, |ne|) tensors
         es, rate_e = self.gimme('energy_spectrum', data_tensor=data_tensor, ptensor=ptensor)
+
         mean_q_produced = (
                 es
                 * self.gimme('lindhard_l', bonus_arg=es,
@@ -797,24 +829,32 @@ class WIMPSource(NRSource):
     """NRSource with time dependent energy spectra from
     wimprates.
     """
-    extra_needed_columns = tuple(
-        list(NRSource.extra_needed_columns)
-        + ['t', 'event_time'])
     # Recoil energies and Wimprates settings
     es = np.geomspace(0.7, 50, 100)  # [keV]
     mw = 1e3  # GeV
     sigma_nucleon = 1e-45  # cm^2
 
-    # Interpolator settings
-    n_in = 10  # Number of time bin edges (wimprates function evaluations + 1)
+    def ignore_columns(self):
+        return super().ignore_columns() + ['wimp_energies']
+
+    n_time_bins = 24
+
+    # If set to True, the energy spectrum at each time will be set to its
+    # average over the data taking period.
+    pretend_wimps_dont_modulate = False
 
     def __init__(self, *args, wimp_kwargs=None, **kwargs):
         # Compute the energy spectrum in a given time range
         # Times used by wimprates are J2000 timestamps
-        assert self.n_in > 1, \
-            f"Number of time bin edges needs to be at least 2"
-        times = np.linspace(wr.j2000(date=self.t_start),
-                            wr.j2000(date=self.t_stop), self.n_in)
+        assert self.n_time_bins >= 1, "Need >= 1 time bin"
+        if hasattr(self, 'n_in'):
+            raise RuntimeError(
+                "n_in is gone! Use n_time_bins to control accuracy, or set "
+                "pretend_wimps_dont_modulate to use a time-averaged spectrum.")
+
+        times = np.linspace(wr.j2000(self.t_start.value),
+                            wr.j2000(self.t_stop.value),
+                            self.n_time_bins + 1)
         time_centers = self.bin_centers(times)
 
         if wimp_kwargs is None:
@@ -843,67 +883,79 @@ class WIMPSource(NRSource):
 
         self.energy_hist = Histdd.from_histogram(spectra,
                                                  bin_edges=(times, es))
+
+        if self.pretend_wimps_dont_modulate:
+            self.energy_hist.histogram = (
+                np.ones_like(self.energy_hist.histogram)
+                * self.energy_hist.sum(axis=0).histogram.reshape(1, -1)
+                / self.n_time_bins)
+
         # Initialize the rest of the source, needs to be after energy_hist is
         # computed because of _populate_tensor_cache
         super().__init__(*args, **kwargs)
 
     def mu_before_efficiencies(self, **params):
-        return self.energy_hist.n / (self.n_in - 1)
+        return self.energy_hist.n / self.n_time_bins
 
     @staticmethod
     def bin_centers(x):
         return 0.5 * (x[1:] + x[:-1])
 
-    def to_event_time(self, jtimes):
-        j_start = wr.j2000(date=self.t_start)
-        j_stop = wr.j2000(date=self.t_stop)
-        assert j_start < j_stop
-
-        ev_time_start = pd.Timestamp(self.t_start).value
-        ev_time_stop = pd.Timestamp(self.t_stop).value
-        assert ev_time_start < ev_time_stop
-
-        jfrac = (jtimes - j_start)/(j_stop - j_start)
-        return jfrac * (ev_time_stop - ev_time_start) + ev_time_start
+    def _batch_data_tensor_shape(self):
+        batch_size, n_names = super()._batch_data_tensor_shape()
+        return [batch_size,
+                n_names + len(self.energy_hist.bin_centers(1))]
 
     def _populate_tensor_cache(self):
         super()._populate_tensor_cache()
-        # Get energy bin centers
+
+        # Create an (n_time_bins, len(es)) histogram of spectra
         e_bin_centers = self.energy_hist.bin_centers(axis=1)
-        # Construct the energy spectra at event times
         e = np.array([self.energy_hist.slicesum(t).histogram
-                      for t in self.data['t']])
+                      for t in self.data['t_j2000']])
+
+        # Look up in which time row/bin each event falls, and concatenate
+        # the expected WIMP energy spectrum to the data tensor.
+        # We modified _fetch so we can access these as 'wimp_energies'
         energy_tensor = tf.convert_to_tensor(e, dtype=fd.float_type())
         assert energy_tensor.shape == [len(self.data), len(e_bin_centers)], \
             f"{energy_tensor.shape} != {len(self.data)}, {len(e_bin_centers)}"
-        self.energy_tensor = tf.reshape(energy_tensor,
-                                        [self.n_batches, self.batch_size, -1])
+        energy_tensor = tf.reshape(energy_tensor,
+                                   [self.n_batches, self.batch_size, -1])
+        self.data_tensor = tf.concat([self.data_tensor,
+                                      energy_tensor],
+                                     axis=2)
 
+        # Store the centers of energy bins separately, these are the same
+        # in each batch.
         es_centers = tf.convert_to_tensor(e_bin_centers,
                                           dtype=fd.float_type())
-        self.all_es_centers = fd.repeat(es_centers[o, :],
-                                        repeats=self.batch_size,
-                                        axis=0)
+        self.es_centers_batch = fd.repeat(es_centers[o, :],
+                                          repeats=self.batch_size,
+                                          axis=0)
+
+    def energy_spectrum(self, wimp_energies):
+        """Return (energies in keV, events at these energies)
+        :param wimp_energies: Expected WIMP energy spectrum at the observed
+        time of each event.
+        """
+        return self.es_centers_batch, wimp_energies
+
+    def _fetch(self, x, data_tensor=None):
+        if x == 'wimp_energies':
+            return data_tensor[:, len(self.name_id):]
+        return super()._fetch(x, data_tensor=data_tensor)
 
     def add_extra_columns(self, d):
         super().add_extra_columns(d)
         # Add J2000 timestamps to data for use with wimprates
-        if 't' not in d:
-            d['t'] = [wr.j2000(date=t)
-                      for t in pd.to_datetime(d['event_time'])]
-
-    def energy_spectrum(self, i_batch):
-        """Return (energies in keV, events at these energies)
-        """
-        batch = tf.dtypes.cast(i_batch[0], dtype=fd.int_type())
-        return self.all_es_centers, self.energy_tensor[batch, :, :]
+        d['t_j2000'] = wr.j2000(d['event_time'])
 
     def _add_random_energies(self, data, n_events):
         """Draw n_events random energies and times from the energy/
         time spectrum and add them to the data dict.
         """
         events = self.energy_hist.get_random(n_events)
-        data['t'] = j2000_times = events[:, 0]
         data['energy'] = events[:, 1]
-        data['event_time'] = self.to_event_time(j2000_times)
+        data['event_time'] = fd.j2000_to_event_time(events[:, 0])
         return data
