@@ -15,103 +15,26 @@ o = tf.newaxis
 
 
 @export
-class SourceBase:
-    """Base class of Source"""
-
+class Source:
+    n_batches = None
+    n_padding = None
     trace_difrate = True
 
-    def _init_padding(self, batch_size, _skip_tf_init):
-        # Annotate requests n_events, currently no padding
-        self.n_padding = 0
-        self.n_events = len(self.data)
-
-        if hasattr(self, 'batch_size'):
-            assert batch_size is None
-        else:
-            if batch_size is None or batch_size > self.n_events or _skip_tf_init:
-                batch_size = self.n_events
-            self.batch_size = max(1, batch_size)
-
-        self.n_batches = np.ceil(
-            self.n_events / self.batch_size).astype(np.int)
-
-        if not _skip_tf_init:
-            # Extend dataframe with events to nearest batch_size multiple
-            # We're using actual events for padding, since using zeros or
-            # nans caused problems with gradient calculation
-            # padded events are clipped when summing likelihood terms
-            self.n_padding = self.n_batches * self.batch_size - len(self.data)
-            if self.n_padding > 0:
-                df_pad = self.data.iloc[:self.n_padding, :]
-                self.data = pd.concat([self.data, df_pad], ignore_index=True)
-
-        # Add i_batch column to data for use with precomputed model functions
-        self.data['i_batch'] = np.repeat(np.arange(self.n_batches),
-                                         self.batch_size)[:len(self.data)]
-
-
-@export
-class ColumnSource(SourceBase):
-    """Source with a fixed mu (specified as self.mu)
-     and differential rate specified by a column in the data (self.column)
-    """
-    column = "Rename_me!"
-    mu = 42.
-
-    trace_difrate = False
-
-    def __init__(self,
-                 data=None,
-                 batch_size=10,
-                 max_sigma=3,
-                 data_is_annotated=False,
-                 _skip_tf_init=False,
-                 _skip_bounds_computation=False,
-                 fit_params=None,
-                 **params):
-        """
-
-        :param data:
-        :param batch_size: used
-        :param max_sigma:
-        :param data_is_annotated:
-        :param _skip_tf_init:
-        :param _skip_bounds_computation:
-        :param fit_params: List of parameters to fit
-        :param params: New defaults
-        """
-        self.defaults = dict()
-        self.data = data
-        if data is None:
-            # We're calling the source without data. Set the batch_size here
-            # since we can't pass it to set_data later
-            self.batch_size = batch_size
-        else:
-            self._init_padding(batch_size, _skip_tf_init)
-            self.data_tensor = fd.np_to_tf(self.data[self.column])
-            self.data_tensor = tf.reshape(self.data_tensor,
-                                          (-1,self.batch_size, 1))
-
-    def differential_rate(self, data_tensor, **params):
-        return data_tensor[:, 0]
-
-    @classmethod
-    def mu_function(cls,
-                    interpolation_method='star',
-                    n_trials=int(1e5),
-                    **params):
-        """Return function that maps params -> expected number of events
-        Parameters must be specified as kwarg=(start, stop, n_anchors)
-        """
-        return lambda **kwargs: cls.mu
-
-
-@export
-class Source(SourceBase):
     data_methods = tuple()
     special_data_methods = tuple()
     inner_dimensions = tuple()
-    extra_needed_columns = tuple()
+
+    # List all columns that are manually _fetch ed here
+    # These will be added to the data_tensor even when the model function
+    # inspection will not find them.
+    def extra_needed_columns(self):
+        return []
+
+    # List all columns for which sneaky hacks are used to intercept _fetch here
+    # These will not be added to the data tensor even when the model function
+    # inspection does find them.
+    def ignore_columns(self):
+        return []
 
     data = None
 
@@ -119,6 +42,39 @@ class Source(SourceBase):
     # Initialization and helpers
     ##
 
+    @classmethod
+    def find_defaults(cls):
+        """Discover which functions need which arguments / dimensions
+        Discover possible parameters.
+        Returns f_dims, f_params and defaults.
+        """
+        f_dims = {x: [] for x in cls.data_methods}
+        f_params = {x: [] for x in cls.data_methods}
+        defaults = dict()
+        for fname in cls.data_methods:
+            f = getattr(cls, fname)
+            if not callable(f):
+                # Constant
+                continue
+            seen_special = False
+            for pname, p in inspect.signature(f).parameters.items():
+                if pname == 'self':
+                    continue
+                if p.default is inspect.Parameter.empty:
+                    if fname in cls.special_data_methods and not seen_special:
+                        seen_special = True
+                    else:
+                        # It's an observable dimension
+                        f_dims[fname].append(pname)
+                else:
+                    # It's a parameter that can be fitted
+                    f_params[fname].append(pname)
+                    if (pname in defaults and p.default != defaults[pname]):
+                        raise ValueError(f"Inconsistent defaults for {pname}")
+                    defaults[pname] = tf.convert_to_tensor(
+                        p.default, dtype=fd.float_type())
+        return f_dims, f_params, defaults
+
     def __init__(self,
                  data=None,
                  batch_size=10,
@@ -142,115 +98,118 @@ class Source(SourceBase):
         self.max_sigma = max_sigma
 
         # Discover which functions need which arguments / dimensions
-        # Discover possible parameters
-        self.f_dims = {x: [] for x in self.data_methods}
-        self.f_params = {x: [] for x in self.data_methods}
-        self.defaults = dict()
-        for fname in self.data_methods:
-            f = getattr(self, fname)
-            if not callable(f):
-                # Constant
-                continue
-            for i, (pname, p) in enumerate(
-                    inspect.signature(f).parameters.items()):
-                if pname == 'i_batch':
-                    # This function uses precomputed data
-                    self.f_dims[fname].append(pname)
-                    continue
-                if p.default is inspect.Parameter.empty:
-                    if not (fname in self.special_data_methods and i == 0):
-                        # It's an observable dimension
-                        self.f_dims[fname].append(pname)
-                else:
-                    # It's a parameter that can be fitted
-                    self.f_params[fname].append(pname)
-                    if (pname in self.defaults
-                            and p.default != self.defaults[pname]):
-                        raise ValueError(f"Inconsistent defaults for {pname}")
-                    self.defaults[pname] = tf.convert_to_tensor(
-                        p.default, dtype=fd.float_type())
+        # Discover possible parameters.
+        self.f_dims, self.f_params, self.defaults = self.find_defaults()
+
+        # Which columns are needed from data?
+        ctc = list(set(sum(self.f_dims.values(), [])))
+        ctc += self.extra_needed_columns()
+        ctc += [x + '_min' for x in self.inner_dimensions]  # Needed in domain
+        ctc = [x for x in ctc if x not in self.ignore_columns()]
+        self.cols_to_cache = ctc
+
+        # Check for duplicate columns and give error
+        _seen = set()
+        for x in self.cols_to_cache:
+            if x in _seen:
+                raise RuntimeError(
+                    f"Column {x} requested twice for data_tensor!")
+            _seen.add(x)
+
+        self.name_id = fd.index_lookup_dict(ctc)
 
         self.set_defaults(**params)
 
         if fit_params is None:
             fit_params = list(self.defaults.keys())
-        self.fit_params = [tf.constant(x) for x in fit_params
+        self.fit_params = [x for x in fit_params
                            if x in self.defaults]
 
-        if len(self.defaults):
-            self.param_id = tf.lookup.StaticVocabularyTable(
-                tf.lookup.KeyValueTensorInitializer(tf.constant(list(self.defaults.keys())),
-                                                    tf.range(len(self.defaults),
-                                                             dtype=tf.dtypes.int64)),
-                num_oov_buckets=1,
-                lookup_key_dtype=tf.dtypes.string)
-            # Indices of params we actually want to fit; we have to differentiate wrt these
-            self.fit_param_indices = tuple([
-                self.param_id.lookup(param_name)
-                for param_name in self.fit_params])
+        self.param_id = fd.index_lookup_dict(self.defaults.keys())
+        # Indices of params we actually want to fit; we have to differentiate wrt these
+        self.fit_param_indices = tuple([
+            self.param_id[param_name]
+            for param_name in self.fit_params])
 
         if data is None:
             # We're calling the source without data. Set the batch_size here
             # since we can't pass it to set_data later
             self.batch_size = batch_size
         else:
+            self.batch_size = min(batch_size, len(data))
             self.set_data(data,
-                          batch_size=batch_size,
                           data_is_annotated=data_is_annotated,
                           _skip_tf_init=_skip_tf_init,
                           _skip_bounds_computation=_skip_bounds_computation)
 
+        if not _skip_tf_init:
+            self.trace_differential_rate()
+
     def set_defaults(self, **params):
         for k, v in params.items():
             if k in self.defaults:
                 self.defaults[k] = tf.convert_to_tensor(
                     v, dtype=fd.float_type())
-            else:
-                raise ValueError(f"Key {k} not in defaults")
 
     def set_data(self,
-                 data,
-                 batch_size=None,
+                 data=None,
                  data_is_annotated=False,
                  _skip_tf_init=False,
                  _skip_bounds_computation=False,
                  **params):
+        self.set_defaults(**params)
+
+        if data is None:
+            self.data = self.n_batches = self.n_padding = None
+            return
         self.data = data
         del data
 
-        self.set_defaults(**params)
+        # Annotate requests n_events, currently no padding
+        self.n_padding = 0
+        self.n_events = len(self.data)
+        self.n_batches = np.ceil(
+            self.n_events / self.batch_size).astype(np.int)
 
-        self._init_padding(batch_size, _skip_tf_init)
+        if not _skip_tf_init:
+            # Extend dataframe with events to nearest batch_size multiple
+            # We're using actual events for padding, since using zeros or
+            # nans caused problems with gradient calculation
+            # padded events are clipped when summing likelihood terms
+            self.n_padding = self.n_batches * self.batch_size - len(self.data)
+            if self.n_padding > 0:
+                # Repeat first event n_padding times and concat to rest of data
+                df_pad = self.data.iloc[np.zeros(self.n_padding)]
+                self.data = pd.concat([self.data, df_pad], ignore_index=True)
 
         if not data_is_annotated:
             self.add_extra_columns(self.data)
             self._annotate(_skip_bounds_computation=_skip_bounds_computation)
 
         if not _skip_tf_init:
+            self._check_data()
             self._populate_tensor_cache()
             self._calculate_dimsizes()
 
-            self.trace_differential_rate()
+    def _check_data(self):
+        """Do any final checks on the self.data dataframe,
+        before passing it on to the tensorflow layer.
+        """
+        for column in self.cols_to_cache:
+            if column not in self.data.columns:
+                raise ValueError(f"Data lacks required column {column}; "
+                                 f"did annotation happen correctly?")
 
     def _populate_tensor_cache(self):
-        # Cache only float and int cols
-        cols_to_cache = [x for x in self.data.columns
-                         if fd.is_numpy_number(self.data[x])]
-
-        self.name_id = tf.lookup.StaticVocabularyTable(
-            tf.lookup.KeyValueTensorInitializer(tf.constant(cols_to_cache),
-                                                tf.range(len(cols_to_cache),
-                                                         dtype=tf.dtypes.int64)),
-            num_oov_buckets=1,
-            lookup_key_dtype=tf.dtypes.string)
 
         # Create one big data tensor (n_batches, events_per_batch, n_cols)
         # TODO: make a list
-        self.data_tensor = tf.constant(self.data[cols_to_cache].values,
+        ctc = self.cols_to_cache
+        self.data_tensor = tf.constant(self.data[ctc].values,
                                        dtype=fd.float_type())
         self.data_tensor = tf.reshape(self.data_tensor, [self.n_batches,
                                                          -1,
-                                                         len(cols_to_cache)])
+                                                         len(ctc)])
 
     def _calculate_dimsizes(self):
         self.dimsizes = dict()
@@ -260,20 +219,29 @@ class Source(SourceBase):
             self.dimsizes[var] = int(tf.reduce_max(ma - mi + 1).numpy())
 
     @contextmanager
-    def _set_temporarily(self, data=None, **kwargs):
-        """Set data and/or defaults temporarily"""
+    def _set_temporarily(self, data, **kwargs):
+        """Set data and/or defaults temporarily, without affecting the
+        data tensor state"""
+        if data is None:
+            raise ValueError("No point in setting data = None temporarily")
         old_defaults = self.defaults
         if data is None:
             self.set_defaults(**kwargs)
         else:
-            old_data = self.data
+            if self.data is None:
+                old_data = None
+            else:
+                old_data = self.data[:self.n_events]  # Remove padding
             self.set_data(data, **kwargs, _skip_tf_init=True)
         try:
             yield
         finally:
             self.defaults = old_defaults
-            if data is not None:
-                self.data = old_data
+            if old_data is not None:
+                self.set_data(
+                    old_data,
+                    data_is_annotated=True,
+                    _skip_tf_init=True)
 
     def annotate_data(self, data, _skip_bounds_computation=False, **params):
         """Add columns to data with inference information"""
@@ -290,21 +258,19 @@ class Source(SourceBase):
         :param x: column name
         :param data_tensor: Data tensor, columns as in self.name_id
         """
+        if x in self.ignore_columns():
+            raise RuntimeError(
+                "Attempt to fetch %s, which is in ignore_columns" % x)
         if data_tensor is None:
             # We're inside annotate, just return the column
             return fd.np_to_tf(self.data[x].values)
 
-        col_id = tf.dtypes.cast(self.name_id.lookup(tf.constant(x)),
-                                fd.int_type())
-        # if i_batch is None:
-        #     return tf.reshape(self.data_tensor[:,:,col_id], [-1])
-        # else:
-        return data_tensor[:, col_id]
+        return data_tensor[:, self.name_id[x]]
 
     def _fetch_param(self, param, ptensor):
         if ptensor is None:
             return self.defaults[param]
-        id = tf.dtypes.cast(self.param_id.lookup(tf.constant(param)),
+        id = tf.dtypes.cast(self.param_id[param],
                             dtype=fd.int_type())
         return ptensor[id]
 
@@ -324,7 +290,6 @@ class Source(SourceBase):
         Before using gimme, you must use set_data to
         populate the internal caches.
         """
-        # TODO: make a clean way to keep track of i_batch or have it as input
         assert (bonus_arg is not None) == (fname in self.special_data_methods)
 
         if data_tensor is None:
@@ -365,18 +330,22 @@ class Source(SourceBase):
     # TODO: remove duplication for batch loop? Also in inference
     def batched_differential_rate(self, progress=True, **params):
         progress = (lambda x: x) if not progress else tqdm
-        y = np.concatenate([
-            fd.tf_to_np(self.differential_rate(
-                data_tensor=self.data_tensor[i_batch],
-                **params))
-            for i_batch in progress(range(self.n_batches))])
-        return y[:self.n_events]
+        y = []
+        for i_batch in progress(range(self.n_batches)):
+            q = self.data_tensor[i_batch]
+            y.append(fd.tf_to_np(self.differential_rate(data_tensor=q,
+                                                        **params)))
+
+        return np.concatenate(y)[:self.n_events]
+
+    def _batch_data_tensor_shape(self):
+        return [self.batch_size, len(self.name_id)]
 
     def trace_differential_rate(self):
         input_signature = (
-            tf.TensorSpec(shape=self.data_tensor.shape[1:],
+            tf.TensorSpec(shape=self._batch_data_tensor_shape(),
                           dtype=fd.float_type()),
-            tf.TensorSpec(shape=[len(self.defaults)],
+            tf.TensorSpec(shape=[len(self.param_id)],
                           dtype=fd.float_type()))
         self._differential_rate_tf = tf.function(
             self._differential_rate,
@@ -425,14 +394,13 @@ class Source(SourceBase):
     # Simulation methods and helpers
     ##
 
-    def simulate(self, energies, fix_truth=None, **params):
-        """Simulate events at energies.
+    def simulate(self, n_events, fix_truth=None, **params):
+        """Simulate n events.
 
-        Will not return | energies | events lost due to
-        selection/detection efficiencies
+        Will not return events lost due to selection/detection efficiencies
         """
         # Draw random "deep truth" variables (energy, position)
-        sim_data = self.random_truth(energies, fix_truth=fix_truth, **params)
+        sim_data = self.random_truth(n_events, fix_truth=fix_truth, **params)
 
         with self._set_temporarily(sim_data, _skip_bounds_computation=True,
                                    **params):
@@ -467,13 +435,17 @@ class Source(SourceBase):
         # Estimate mus under the specified variations
         pspaces = dict()    # parameter -> tf.linspace of anchors
         mus = dict()        # parameter -> tensor of mus
-        for pname, pspace_spec in tqdm(param_specs.items(),
+        for pname, (start, stop, n) in tqdm(param_specs.items(),
                                        desc="Estimating mus"):
-            pspaces[pname] = tf.linspace(*pspace_spec)
+            # Parameters are floats, but users might input ints as anchors
+            # accidentally, triggering a confusing tensorflow device placement
+            # message
+            start, stop = float(start), float(stop)
+            pspaces[pname] = tf.linspace(start, stop, n)
             mus[pname] = tf.convert_to_tensor(
                  [self.estimate_mu(**{pname: x}, n_trials=n_trials)
-                  for x in np.linspace(*pspace_spec)],
-                dtype=fd.float_type())
+                  for x in np.linspace(start, stop, n)],
+                 dtype=fd.float_type())
 
         def mu_itp(**kwargs):
             mu = base_mu
@@ -487,12 +459,6 @@ class Source(SourceBase):
 
         return mu_itp
 
-    def mu_before_efficiencies(self, **params):
-        """Return mean expected number of events BEFORE efficiencies/response
-        using data for the evaluation of the energy spectra
-        """
-        raise NotImplementedError
-
     def estimate_mu(self, n_trials=int(1e5), **params):
         """Return estimate of total expected number of events
         :param n_trials: Number of events to simulate for estimate
@@ -502,14 +468,22 @@ class Source(SourceBase):
                 * len(d_simulated) / n_trials)
 
     ##
-    # Functions probably want to override
+    # Functions you have to override
     ##
-    def energy_spectrum(self, **params):
-        raise NotImplementedError
 
     def _differential_rate(self, data_tensor, ptensor):
         raise NotImplementedError
 
+    def mu_before_efficiencies(self, **params):
+        """Return mean expected number of events BEFORE efficiencies/response
+        using data for the evaluation of the energy spectra
+        """
+        raise NotImplementedError
+
+    ##
+    # Functions you probably should override
+    ##
+
     def _annotate(self, _skip_bounds_computation=False):
         """Add columns needed in inference to self.data
         :param _skip_bounds_computation: Do not compute min/max bounds
@@ -520,9 +494,6 @@ class Source(SourceBase):
     def add_extra_columns(self, data):
         """Add additional columns to data
 
-        You must add any columns from data you use here to
-        extra_needed.columns.
-
         :param data: pandas DataFrame
         """
         pass
@@ -535,4 +506,29 @@ class Source(SourceBase):
 
     def _simulate_response(self):
         """Do a forward simulation of the detector response, using self.data"""
-        raise NotImplementedError
+        return self.data
+
+
+@export
+class ColumnSource(Source):
+    """Source that expects precomputed differential rate in a column,
+    and precomputed mu in an attribute
+    """
+    column = 'rename_me!'
+    mu = 42.
+
+    def extra_needed_columns(self):
+        return super().extra_needed_columns() + [self.column]
+
+    def estimate_mu(self, n_trials=None, **params):
+        return self.mu
+
+    def mu_before_efficiencies(self, **params):
+        return self.mu
+
+    def _differential_rate(self, data_tensor, ptensor):
+        return self._fetch(self.column, data_tensor)
+
+    def random_truth(self, n_events, fix_truth=None, **params):
+        print(f"{self.__class__.__name__} cannot generate events, skipping")
+        return pd.DataFrame()
